(window.webpackJsonp=window.webpackJsonp||[]).push([[288],{1468:function(n,i){n.exports={attributes:{title:"OpenAI",description:"The ArgillaOpenAITrainer leverages the features of OpenAI to fine-tune programmatically with Argilla.",links:[{linkText:"Argilla docs",linkLink:"https://docs.argilla.io/en/latest/practical_guides/fine_tune.html#text-classification"},{linkText:"OpenAI docs",linkLink:"https://platform.openai.com/docs/guides/fine-tuning"}]},body:'*code snippet*\n\n```python\nfrom argilla.training import ArgillaTrainer\n\ntrainer = ArgillaTrainer(\n    name="<my_dataset_name>",\n    workspace="<my_workspace_name>",\n    framework="setfit",\n    train_size=0.8\n)\ntrainer.update_config(num_iterations=10)\ntrainer.train(output_dir="token-classification")\nrecords = trainer.predict("The ArgillaTrainer is great!", as_argilla_records=True)\n```\n\n*update training config*\n\n```python\ntrainer.update_config(\n    training_file = None,\n    validation_file = None,\n    model = "curie,\n    n_epochs = 2,\n    batch_size = None,\n    learning_rate_multiplier = 0.1,\n    prompt_loss_weight = 0.1,\n    compute_classification_metrics = False,\n    classification_n_classes = None,\n    classification_positive_class = None,\n    classification_betas = None,\n    suffix = None\n)\n```\n'}}}]);