(window.webpackJsonp=window.webpackJsonp||[]).push([[254],{1689:function(e,t,i){"use strict";i.r(t),t.default={multi_label_selection:"Multi-label",ranking:"Ranking",label_selection:"Label",text:"Text",rating:"Rating",minimize:"Minimize",select:"Select",search:"Search",expand:"Expand",copied:"Copied",title:"Title",description:"Description",useMarkdown:"Use Markdown",visibleForAnnotators:"Visible for annotators",allowExtraMetadata:"Allow extra metadata",extraMetadata:"Extra metadata",dimension:"Dimension",visibleOptions:"Visible options",annotationGuidelines:"Annotation guidelines",noAnnotationGuidelines:"This dataset has no annotation guidelines",breadcrumbs:{home:"Home",datasetSettings:"settings",userSettings:"my settings"},userSettings:{title:"My settings"},settings:{title:"Dataset settings",seeYourDataset:"See your dataset",editFields:"Edit fields",editQuestions:"Edit questions",editMetadata:"Edit metadata properties",editVectors:"Edit vectors",deleteDataset:"Delete dataset",deleteWarning:"Be careful, this action is not reversible",deleteConfirmation:"Delete confirmation",deleteConfirmationMessage:"You are about to delete: <strong>{datasetName}</strong> from workspace <strong>{workspaceName}</strong>. This action cannot be undone",yesDelete:"Yes, delete"},button:{ignore_and_continue:"Ignore and continue",login:"Sign in","hf-login":"Sign in with Hugging Face",sign_in_with_username:"Sign in with username",cancel:"Cancel",continue:"Continue"},to_submit_complete_required:"To submit complete \nrequired responses",some_records_failed_to_annotate:"Some records failed to annotate",changes_no_submit:"You didn't submit your changes",bulkAnnotation:{recordsSelected:"1 record selected | {count} records selected",recordsViewSettings:"Record size",fixedHeight:"Collapse records",defaultHeight:"Expand records",to_annotate_record_bulk_required:"No record selected",select_to_annotate:"Select all",pageSize:"Page size",selectAllResults:"Select all {total} matched records",haveSelectedRecords:"You have selected all {total} records",actionConfirmation:"Bulk action confirmation",actionConfirmationText:"This action will affect {total} records, do you want to continue? ",allRecordsAnnotated:"The {total} records have been {action}",affectedAll:{submitted:"submitted",discarded:"discarded",draft:"saved as draft"}},shortcuts:{label:"Shortcuts",pagination:{go_to_previous_record:"Previous (←)",go_to_next_record:"Next (→)"}},questions_form:{discard:"Discard",submit:"Submit",draft:"Save as draft"},sorting:{addOtherField:"+ Add another field",suggestion:{score:"Suggestion score",value:"Suggestion value"},response:"Response value",record:"general",metadata:"metadata"},suggestion:{filter:{value:"Suggestion values",score:"Score",agent:"Agent"},plural:"Suggestions","suggested-rank":"Suggested rank",name:"Suggestion"},similarity:{"record-number":"Record number",findSimilar:"Find similar",similarTo:"Similar to",similarityScore:"Similarity Score",similarUsing:"similar using",expand:"Expand",collapse:"Collapse"},login:{title:"Sign in",claim:"Work on data together.</br>Make your models better.",support:"To get support from the community, join us on <a href='{link}' target='_blank'>Slack</a>",quickstart:"You are using the Quickstart version of Argilla. Check <a href='{link}' target='_blank'>this guide</a> to learn more about usage and configuration options.",hf:{title:"Welcome to {space}",subtitle:"Join <strong>{user}</strong> to build better datasets for AI"}},status:"Status",filters:"Filters",filterBy:"Filter by...",fields:"Fields",questions:"Questions",metadata:"Metadata",vectors:"Vectors",dangerZone:"Danger zone",responses:"Responses","reset-all":"Reset all",reset:"Reset",less:"Less",with:"with",find:"Find",cancel:"Cancel",focus_mode:"Focus view",bulk_mode:"Bulk view",update:"Update",youAreOnlineAgain:"You are online again",youAreOffline:"You are offline",argilla:{api:{"errors::UnauthorizedError":"Could not validate credentials"}}}}}]);