(window.webpackJsonp=window.webpackJsonp||[]).push([[259],{1207:function(a,e,t){"use strict";t.r(e),t.d(e,"useDatasetViewModel",(function(){return c}));t(15);var n=t(73),r=t(44),o=t(292),s=t(239),c=function useDatasetViewModel(){var a=Object(n.d)(!1),e=Object(n.h)(),t=Object(s.i)(),c=e.value.params.id;return{datasetId:c,isLoadingDataset:a,handleError:function handleError(a){var e="";switch(a){case o.a.ERROR_FETCHING_DATASET_INFO:e="Can't get dataset info for dataset_id: ".concat(c);break;case o.a.ERROR_FETCHING_WORKSPACE_INFO:e="Can't get workspace info for dataset_id: ".concat(c);break;default:e="There was an error on fetching dataset info and workspace info. Please try again"}var t={message:e,numberOfChars:e.length,type:"error"};r.a.dispatch("notify",t)},createRootBreadCrumbs:function createRootBreadCrumbs(a){return[{link:{name:"datasets"},name:t("breadcrumbs.home")},{link:{path:"/datasets?workspaces=".concat(a.workspace)},name:a.workspace},{link:{path:"annotation-mode"},name:a.name}]}}}}}]);