(window.webpackJsonp=window.webpackJsonp||[]).push([[263],{1477:function(e,t,n){"use strict";n.r(t),n.d(t,"useDatasetsViewModel",(function(){return i}));n(10),n(6);var r=n(20),a=n(73),c=n(309),o=n(307),u=n(239),__awaiter=function(e,t,n,r){return new(n||(n=Promise))((function(a,c){function fulfilled(e){try{step(r.next(e))}catch(e){c(e)}}function rejected(e){try{step(r.throw(e))}catch(e){c(e)}}function step(e){e.done?a(e.value):function adopt(e){return e instanceof n?e:new n((function(t){t(e)}))}(e.value).then(fulfilled,rejected)}step((r=r.apply(e,t||[])).next())}))},i=function useDatasetsViewModel(){var e=Object(a.d)(!1),t=Object(o.a)().state,n=Object(r.useResolve)(c.a);Object(u.e)((function(){return __awaiter(void 0,void 0,void 0,regeneratorRuntime.mark((function _callee(){return regeneratorRuntime.wrap((function _callee$(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,i();case 2:case"end":return e.stop()}}),_callee)})))})),Object(a.g)((function(){s()}));var i=function onLoadDatasets(){return __awaiter(void 0,void 0,void 0,regeneratorRuntime.mark((function _callee2(){return regeneratorRuntime.wrap((function _callee2$(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,n.execute();case 2:case"end":return e.stop()}}),_callee2)})))},s=function loadDatasets(){return __awaiter(void 0,void 0,void 0,regeneratorRuntime.mark((function _callee3(){return regeneratorRuntime.wrap((function _callee3$(t){for(;;)switch(t.prev=t.next){case 0:return e.value=!0,t.next=3,i();case 3:e.value=!1;case 4:case"end":return t.stop()}}),_callee3)})))};return{datasets:t,isLoadingDatasets:e}}}}]);