#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer


def start(
    host: str = typer.Option("0.0.0.0", help="The host where the Argilla server will be binded"),
    port: int = typer.Option(6900, help="The port where the Argilla server will be binded"),
    access_log: bool = typer.Option(True, help="Whether to enable or disable the Argilla server access log"),
) -> None:
    import uvicorn

    uvicorn.run(
        "argilla_server:app",
        port=port,
        host=host,
        access_log=access_log,
    )
