(window.webpackJsonp=window.webpackJsonp||[]).push([[265,266],{1490:function(e,t,n){"use strict";n.r(t),n.d(t,"useOAuthViewModel",(function(){return a}));n(10),n(6);var r=n(73),u=n(20),c=n(44),o=n(199),i=n(239),__awaiter=function(e,t,n,r){return new(n||(n=Promise))((function(u,c){function fulfilled(e){try{step(r.next(e))}catch(e){c(e)}}function rejected(e){try{step(r.throw(e))}catch(e){c(e)}}function step(e){e.done?u(e.value):function adopt(e){return e instanceof n?e:new n((function(t){t(e)}))}(e.value).then(fulfilled,rejected)}step((r=r.apply(e,t||[])).next())}))},a=function useOAuthViewModel(){var e=Object(i.i)(),t=Object(r.h)(),n=Object(i.h)(),a=Object(u.useResolve)(o.a);Object(r.g)((function(){return __awaiter(void 0,void 0,void 0,regeneratorRuntime.mark((function _callee(){return regeneratorRuntime.wrap((function _callee$(e){for(;;)switch(e.prev=e.next){case 0:return e.next=2,s();case 2:case"end":return e.stop()}}),_callee)})))}));var s=function tryLogin(){return __awaiter(void 0,void 0,void 0,regeneratorRuntime.mark((function _callee2(){var r,u,o,i;return regeneratorRuntime.wrap((function _callee2$(s){for(;;)switch(s.prev=s.next){case 0:return r=t.value,u=r.params,o=r.query,i=u.provider,s.prev=2,s.next=5,a.login(i,o);case 5:s.next=10;break;case 7:s.prev=7,s.t0=s.catch(2),c.a.dispatch("notify",{message:e("argilla.api.errors::UnauthorizedError"),type:"error"});case 10:return s.prev=10,n.go("/"),s.finish(10);case 13:case"end":return s.stop()}}),_callee2,null,[[2,7,10,13]])})))}}},1703:function(e,t,n){"use strict";n.r(t);var r=n(1490),u={setup:function setup(){return Object(r.useOAuthViewModel)()}},c=n(38),o=Object(c.a)(u,(function render(){return(0,this._self._c)("div")}),[],!1,null,null,null);t.default=o.exports}}]);