(window.webpackJsonp=window.webpackJsonp||[]).push([[239,184],{1716:function(e,t,r){"use strict";r.r(t);var i={props:{dataset:{type:Object,required:!0},currentMetric:{type:String}},data:function data(){return{sidebarItems:[{id:"annotate",tooltip:"Hand labeling",icon:"hand-labeling",group:"Mode",action:"change-view-mode",relatedMetrics:["progress","stats"]},{id:"explore",tooltip:"Exploration",icon:"exploration",group:"Mode",action:"change-view-mode",relatedMetrics:["progress","stats"]},{id:"progress",tooltip:"Progress",icon:"progress",action:"show-metrics",group:"Metrics"},{id:"stats",tooltip:"Stats",icon:"stats",action:"show-metrics",group:"Metrics"},{id:"refresh",tooltip:"Refresh",icon:"refresh",group:"Refresh",action:"refresh"}]}},beforeMount:function beforeMount(){this.$emit("set-sidebar-items",this.sidebarItems)},methods:{onChangeViewMode:function onChangeViewMode(e){this.$emit("change-view-mode",e)},onShowMetrics:function onShowMetrics(e){this.$emit("show-metrics",e)}}},o=r(38),s=Object(o.a)(i,(function render(){var e=this;return(0,e._self._c)("sidebar-menu",{attrs:{"current-metric":e.currentMetric,dataset:e.dataset,"sidebar-items":e.sidebarItems},on:{refresh:function refresh(t){return e.$emit("refresh")},"show-metrics":e.onShowMetrics,"change-view-mode":e.onChangeViewMode}})}),[],!1,null,null,null);t.default=s.exports;installComponents(s,{SidebarMenu:r(1264).default})},797:function(e,t,r){r(176).register({refresh:{width:40,height:40,viewBox:"0 0 40 40",data:'<path pid="0" d="M33.861 17.47L30.521 5 26.98 8.54C20.5 4.458 11.836 5.238 6.193 10.882l2.635 2.635c4.178-4.178 10.485-4.93 15.429-2.253l-2.865 2.864 12.47 3.341zM6.139 22.53L9.479 35l3.541-3.54c6.48 4.082 15.144 3.302 20.787-2.342l-2.635-2.635c-4.178 4.178-10.485 4.93-15.429 2.253l2.865-2.864-12.47-3.341z" _fill="#000"/>'}})}}]);