(window.webpackJsonp=window.webpackJsonp||[]).push([[268],{1484:function(e,n,u){"use strict";u.r(n),u.d(n,"useWelcomeHFViewModel",(function(){return a}));var i=u(20),g=u(199),o=u(219),a=function useWelcomeHFViewModel(){var e=Object(i.useResolve)(g.a),n=Object(o.a)();return{authorize:function authorize(){e.authorize("huggingface")},isRunningOnHuggingFace:n.isRunningOnHuggingFace,hasHuggingFaceOAuthConfigured:n.hasHuggingFaceOAuthConfigured}}}}]);