#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.daos.models.datasets import BaseDatasetDB
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.errors import MissingDatasetRecordsError


def test_raise_proper_error():
    dao = DatasetRecordsDAO.get_instance(GenericElasticEngineBackend.get_instance())
    with pytest.raises(MissingDatasetRecordsError):
        dao.search_records(
            dataset=BaseDatasetDB(
                name="mock-notfound",
                workspace="workspace-name",
                task=TaskType.text_classification,
            )
        )
