#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture
    from sqlalchemy.ext.asyncio import AsyncSession


@pytest.fixture(autouse=True)
def mock_session_local(mocker: "MockerFixture", async_db_proxy: "AsyncSession") -> None:
    mocker.patch("argilla_server.cli.database.users.create.AsyncSessionLocal", return_value=async_db_proxy)
    mocker.patch("argilla_server.cli.database.users.update.AsyncSessionLocal", return_value=async_db_proxy)
    mocker.patch("argilla_server.cli.database.users.create_default.AsyncSessionLocal", return_value=async_db_proxy)
    mocker.patch("argilla_server.cli.database.users.migrate.AsyncSessionLocal", return_value=async_db_proxy)
