(window.webpackJsonp=window.webpackJsonp||[]).push([[137],{911:function(e,t,n){"use strict";n.r(t);n(28),n(87),n(6),n(42),n(149),n(21),n(88);var i={data:function data(){return{timer:null,keyCode:""}},computed:{options:function options(){var e;return null===(e=this.$slots.default[0].context)||void 0===e?void 0:e.question.answer.values}},methods:{selectByKeyCode:function selectByKeyCode(e){var t=this.options.find((function(t){return t.value==e}));t&&document.getElementById(t.id).click(),this.reset()},hasJustOneCoincidence:function hasJustOneCoincidence(e){return 1==this.options.filter((function(t){return t.value.toString().startsWith(e)})).length},reset:function reset(){this.keyCode="",this.timer=null},answerRatingFor:function answerRatingFor(e){var t=this;if(this.timer&&clearTimeout(this.timer),"Tab"!=e.key)return"Space"==e.code?document.activeElement.click():(this.keyCode+=e.key,isNaN(this.keyCode)?this.reset():this.hasJustOneCoincidence(this.keyCode)?this.selectByKeyCode(this.keyCode):void(this.timer=setTimeout((function(){t.selectByKeyCode(t.keyCode)}),300)))}}},o=n(38),s=Object(o.a)(i,(function render(){var e=this;return(0,e._self._c)("div",{on:{keydown:function keydown(t){return t.preventDefault(),e.answerRatingFor.apply(null,arguments)}}},[e._t("default")],2)}),[],!1,null,null,null);t.default=s.exports}}]);