(window.webpackJsonp=window.webpackJsonp||[]).push([[274],{1473:function(e,o){e.exports={attributes:{},body:"We are not perfect so we love to get feedback.\\\nDo you feel anything is missing in our documentation or do you simply want some help? \\\nFeel free to contact our favorite developer advocate David, via [email](mailto:david@argilla.io), [Slack](https://join.slack.com/t/rubrixworkspace/shared_invite/zt-whigkyjn-a3IUJLD7gDbTZ0rKlvcJ5g), or by scheduling a video meeting in his [Calendly](https://calendly.com/argumentation-at-argilla/30min).\n"}}}]);