(window.webpackJsonp=window.webpackJsonp||[]).push([[275],{1474:function(e,n){e.exports={attributes:{},body:"You can speed up your annotation by using these keyboard shortcuts:\n\n| Action                                 | Keys                              |\n| -------------------------------------- | --------------------------------- |\n| Activate form                          | `⇥ Tab`                           |\n| Show and hide keys in labels (Mac os)  | `⌘ Cmd` (long press)              |\n| Show and hide keys in labels (Other)   | `Ctrl` (long press)               |\n| Move between questions                 | `↓ Down arrow` or `↑ Up arrow`    |\n| Select and unselect label              | `1`, `2`, `3`                     |\n| Move between labels or ranking options | `⇥ Tab` or `⇧ Shift` `⇥ Tab`      |\n| Select rating and rank                 | `1`, `2`, `3`                     |\n| Activate text area                     | `⇧ Shift` `↵ Enter`               |\n| Exit text area                         | `Esc`                             |\n| Discard                                | `⌫ Backspace`                     |\n| Save draft (Mac os)                    | `⌘ Cmd` `S`                       |\n| Save draft (Other)                     | `Ctrl` `S`                        |\n| Submit                                 | `↵ Enter`                         |\n| Move between pages                     | `→ Right arrow` or `← Left arrow` |\n"}}}]);